# -*- coding: utf-8 -*-
"""
exceptions
----------

This module provides the ``CryptoFactoryError`` Exception class.
"""

__all__ = [
    'CryptoFactoryError',
]


class CryptoFactoryError(Exception):
    """
    Crypto-Factory exceptions class to simplify errors management in client interface.
    """
    pass
